#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#ifdef __linux__
#include <glob.h>
#endif
#ifndef __linux__
#include <emx/syscalls.h>
#include <process.h>
#endif

#ifdef UNIX
#define PATH_DELIM "/"
#else
#define PATH_DELIM "\\"
#endif


// ******************************************************
// * FreePoint was created as a quick and dirty hack in *
// * order to easy automatic configuration of FidoNet   *
// * software. You will probably dislike the spaghetti  *
// * code without documentation, but hey, nobody forces *
// * you to make use of this code. :-)                  *
// * The code is released under the GNU PUBLIC LICENSE. *
// * See COPYING for details.                           *
// ******************************************************


// Some hints for compilation:
// The code was written to compile under GCC/EMX.
// If you define __linux__, some changes for portability will
// be used.
// If you define __deutsch, German texts and commands will be
// used.

const unsigned char *CaseIgnore = (unsigned char*)
/*
Standard ASCII sort sequence:
-        |00                 |05                 |10                 |15
-        \x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0A\x0B\x0C\x0D\x0E\x0F
*/      "\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0A\x0B\x0C\x0D\x0E\x0F"
/*
-        |16             |20                 |25                 |30
-        \x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1A\x1B\x1C\x1D\x1E\x1F
*/      "\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1A\x1B\x1C\x1D\x1E\x1F"

/*
-       |32       |40       |50       |60       |70
-        !\"#$%&\'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNO
*/     " !\"#$%&\'()*+,-./0123456789:;<=>?@abcdefghijklmno"
/*
-       |80       |90         |100      |110      |120
-       PQRSTUVWXYZ[\\]^_\'abcdefghijklmnopqrstuvwxyz{|}~
*/     "pqrstuvwxyz[\\]^_\'abcdefghijklmnopqrstuvwxyz{|}~"
/*
-         |130      |140      |150      |160      |170
-       
*/     "a"
/*
-           |180      |190      |200      |210      |220
-       
*/     ""
/*
-             |230      |240      |250
-       \xFF
*/     "\xFF";

/* CCITT Lookup Table */
unsigned short ccitt_table[256] =
{
  /*   0 */  0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50A5, 0x60C6, 0x70E7,
  /*   8 */  0x8108, 0x9129, 0xA14A, 0xB16B, 0xC18C, 0xD1AD, 0xE1CE, 0xF1EF,
  /*  16 */  0x1231, 0x0210, 0x3273, 0x2252, 0x52B5, 0x4294, 0x72F7, 0x62D6,
  /*  24 */  0x9339, 0x8318, 0xB37B, 0xA35A, 0xD3BD, 0xC39C, 0xF3FF, 0xE3DE,
  /*  32 */  0x2462, 0x3443, 0x0420, 0x1401, 0x64E6, 0x74C7, 0x44A4, 0x5485,
  /*  40 */  0xA56A, 0xB54B, 0x8528, 0x9509, 0xE5EE, 0xF5CF, 0xC5AC, 0xD58D,
  /*  48 */  0x3653, 0x2672, 0x1611, 0x0630, 0x76D7, 0x66F6, 0x5695, 0x46B4,
  /*  56 */  0xB75B, 0xA77A, 0x9719, 0x8738, 0xF7DF, 0xE7FE, 0xD79D, 0xC7BC,
  /*  64 */  0x48C4, 0x58E5, 0x6886, 0x78A7, 0x0840, 0x1861, 0x2802, 0x3823,
  /*  72 */  0xC9CC, 0xD9ED, 0xE98E, 0xF9AF, 0x8948, 0x9969, 0xA90A, 0xB92B,
  /*  80 */  0x5AF5, 0x4AD4, 0x7AB7, 0x6A96, 0x1A71, 0x0A50, 0x3A33, 0x2A12,
  /*  88 */  0xDBFD, 0xCBDC, 0xFBBF, 0xEB9E, 0x9B79, 0x8B58, 0xBB3B, 0xAB1A,
  /*  96 */  0x6CA6, 0x7C87, 0x4CE4, 0x5CC5, 0x2C22, 0x3C03, 0x0C60, 0x1C41,
  /* 104 */  0xEDAE, 0xFD8F, 0xCDEC, 0xDDCD, 0xAD2A, 0xBD0B, 0x8D68, 0x9D49,
  /* 112 */  0x7E97, 0x6EB6, 0x5ED5, 0x4EF4, 0x3E13, 0x2E32, 0x1E51, 0x0E70,
  /* 120 */  0xFF9F, 0xEFBE, 0xDFDD, 0xCFFC, 0xBF1B, 0xAF3A, 0x9F59, 0x8F78,
  /* 128 */  0x9188, 0x81A9, 0xB1CA, 0xA1EB, 0xD10C, 0xC12D, 0xF14E, 0xE16F,
  /* 136 */  0x1080, 0x00A1, 0x30C2, 0x20E3, 0x5004, 0x4025, 0x7046, 0x6067,
  /* 144 */  0x83B9, 0x9398, 0xA3FB, 0xB3DA, 0xC33D, 0xD31C, 0xE37F, 0xF35E,
  /* 152 */  0x02B1, 0x1290, 0x22F3, 0x32D2, 0x4235, 0x5214, 0x6277, 0x7256,
  /* 160 */  0xB5EA, 0xA5CB, 0x95A8, 0x8589, 0xF56E, 0xE54F, 0xD52C, 0xC50D,
  /* 168 */  0x34E2, 0x24C3, 0x14A0, 0x0481, 0x7466, 0x6447, 0x5424, 0x4405,
  /* 176 */  0xA7DB, 0xB7FA, 0x8799, 0x97B8, 0xE75F, 0xF77E, 0xC71D, 0xD73C,
  /* 184 */  0x26D3, 0x36F2, 0x0691, 0x16B0, 0x6657, 0x7676, 0x4615, 0x5634,
  /* 192 */  0xD94C, 0xC96D, 0xF90E, 0xE92F, 0x99C8, 0x89E9, 0xB98A, 0xA9AB,
  /* 200 */  0x5844, 0x4865, 0x7806, 0x6827, 0x18C0, 0x08E1, 0x3882, 0x28A3,
  /* 208 */  0xCB7D, 0xDB5C, 0xEB3F, 0xFB1E, 0x8BF9, 0x9BD8, 0xABBB, 0xBB9A,
  /* 216 */  0x4A75, 0x5A54, 0x6A37, 0x7A16, 0x0AF1, 0x1AD0, 0x2AB3, 0x3A92,
  /* 224 */  0xFD2E, 0xED0F, 0xDD6C, 0xCD4D, 0xBDAA, 0xAD8B, 0x9DE8, 0x8DC9,
  /* 232 */  0x7C26, 0x6C07, 0x5C64, 0x4C45, 0x3CA2, 0x2C83, 0x1CE0, 0x0CC1,
  /* 240 */  0xEF1F, 0xFF3E, 0xCF5D, 0xDF7C, 0xAF9B, 0xBFBA, 0x8FD9, 0x9FF8,
  /* 248 */  0x6E17, 0x7E36, 0x4E55, 0x5E74, 0x2E93, 0x3EB2, 0x0ED1, 0x1EF0
};


//#define __deutsch

#ifdef __deutsch

  #define iGNORE           "  Dateiangabe ignoriert"
  #define cANTOPEN         "  Kann Datei nicht ffnen"
  #define bEARBEITEZEILE   "  Bearbeite Zeile '%s'"
  #define kEINEAKTMDATEI   "  - Keine aktuelle Macro-Datei ausgewhlt!"
  #define sUCHBEGRNOTFOUND "  - Konnte Suchbegriff nicht finden"
  #define uMBENENNENERR    "Umbenennen von '%s' nach '%s' fehlgeschlagen."

  #define tRENNUNGLEN  9
  #define sTARTLEN     7
  #define fLAGLEN     10
  #define iFLEN        5

#else

  #define iGNORE           "  file ignored"
  #define cANTOPEN         "  can't open file!"
  #define bEARBEITEZEILE   "  processing line '%s'"
  #define kEINEAKTMDATEI   "  - no current macro file selected!"
  #define sUCHBEGRNOTFOUND "  - could not find key word"
  #define uMBENENNENERR    "error renaming '%s' to '%s'."

  #define tRENNUNGLEN 10
  #define sTARTLEN     6
  #define fLAGLEN      6
  #define iFLEN        3

#endif

short int MAX = 50;

#define MAXLEN 1024
#define MAX 50

char logfile[255] = "freepnt.log";
char VERSION[20] = "FreePoint v2.42";

char *configFile = NULL;
char *macroFile = NULL;

void crc16_init(unsigned short *crcptr) {
    *crcptr = 0x0UL;
}

void crc16_process(unsigned short *crcptr, const unsigned char *buffer, unsigned long length) {
    const unsigned char *ptr = buffer;
    unsigned long ctr;
    int k;

    for (ctr = 0; ctr < length; ctr++, ptr++) {
        *crcptr = (((*crcptr) << 8) ^ ccitt_table [ ((*crcptr) >> 8) ^ (*ptr) ]) & 0xFFFF;
    }
}

void crc16_finalize(unsigned short *crcptr) {}

void monat(char* pmname, int mt) {

   static char* monate[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", \
                             "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
   strcpy(pmname,monate[mt]);
}

void writelog (const char* flag, char* text) {
   static FILE* log;

   struct tm* zt;
   time_t sec = 0;
   char   mt[4];
   char   buffer[32];

   if (log == NULL) {
      if ((log = fopen(logfile, "at")) == NULL) {
#ifdef __deutsch
         printf("\nLogfile kann nicht erstellt werden.");
#else
         printf("\nCannot create logfile.");
#endif
         return;
      }
      fprintf(log, "\n");
   }

   if (strcmp(flag, "0") == 0) {
      fclose(log);
      return;
   }

   sec = time(&sec);
   zt = localtime(&sec);
   monat(mt, zt->tm_mon);
   sprintf(buffer, "\n%s %02i %s %02i:%02i:%02i FPT  ", flag,
           zt->tm_mday, mt, zt->tm_hour, zt->tm_min, zt->tm_sec);
   if (log != NULL) fprintf(log, "%s%s", buffer, text);
}

void mid (char* quelle, int pos, int anz, char* ziel) {
   int idum;

   quelle += pos;
   for (idum = pos; idum < (pos + anz); idum++) {
      *ziel = *quelle;
      quelle++;
      ziel++;
   }
   *ziel = 0;
}

int iinstr (unsigned char *s1, unsigned char *s2, unsigned int pos) {
    unsigned char *si, *sj, *sk;

    if (pos < 1) si = s1;
    else si = s1 + pos;
    while (*si) {
       sj = s2;
       sk = si;
       while (CaseIgnore[*sj] == CaseIgnore[*sk]) {
          sj++;
          if (*sj == 0) return ((int)(si-s1));
          sk++;
       }
       si++;
    }
    return (-1);
}

void rtrim (char* string) {
   char * sptr = string + strlen(string) - 1;

   while(isspace(*sptr)) sptr--;
   *(sptr+1)=0;
}

void ltrim (char* string) {
   char * sptr=string;

   while((*sptr) == 9 || (*sptr) == 32) sptr++;
   strcpy(string, sptr);
}

void underscore(char* text) {
   char* pdumstr = NULL;

   pdumstr = text;
   while (*pdumstr != 0) {
      if (*pdumstr == 32) *pdumstr = 95;
      pdumstr++;
   }
}

void domacro(char* text, char* macro, int macrolen, char* neu) {
   char dumstr[MAXLEN] = "\0";
   char ergebnis[MAXLEN] = "\0";

   int  idum = 0;

   while ((idum = iinstr((unsigned char*)text, (unsigned char*) macro, 0)) != -1) {
        ergebnis[0] = 0;
        if (idum > 0) mid(text, 0, idum, ergebnis);
        strncat(ergebnis, neu, MAXLEN - 1);
        mid(text, idum + macrolen, strlen(text) - idum - macrolen, dumstr);
        strncat(ergebnis, dumstr, MAXLEN - 1);
        strcpy(text, ergebnis);
   }
}

void domacros(char* text, char** pwd, unsigned int nr, char** argstuff) {
   char dumstr[MAXLEN] = "\0";
   char strdum[MAXLEN] = "\0";

   unsigned short crc = 0;

   int idum = 0;
   int idum2 = 0;
   int idum3 = 0;

   unsigned long ldum = 0;

   size_t sdum = 0;

   for (idum = 1; idum < 5; idum++) {
      sprintf(dumstr, "$PWD%i$", idum);
      domacro(text, dumstr, 6, pwd[idum - 1]);
   }

   sprintf(dumstr, "%u", nr);
   domacro(text, "$PNT$", 5, dumstr);

   for (idum = 1; idum < MAX; idum++) {
      sprintf(dumstr, "$N%i$", idum);
      domacro(text, dumstr, strlen(dumstr), argstuff[idum]);
      sprintf(dumstr, "$U%i$", idum);
      strcpy(strdum, argstuff[idum]);
      underscore(strdum);
      domacro(text, dumstr, strlen(dumstr), strdum);
   }

   idum = iinstr((unsigned char*) text, (unsigned char*) "$POINT$(", 0);
   while (idum != -1) {
      idum2 = iinstr((unsigned char*) text, (unsigned char*) ")", idum);
      if (idum2 != -1) {
         mid(text, idum + 8, idum2 - 8 - idum, dumstr);
         rtrim(dumstr);
         idum3 = iinstr((unsigned char*) dumstr, (unsigned char*) ".", 0);
         if (idum3 != -1) mid(dumstr, idum3 + 1, strlen(dumstr) - idum3 - 1, dumstr);
         sscanf(dumstr, "%ld", &ldum);
         sprintf(dumstr, "%ld", ldum);
         mid(text, 0, idum, strdum);
         strncat(strdum, dumstr, MAXLEN);
         mid(text, idum2 + 1, strlen(text) - idum2 - 1, dumstr);
         strncat(strdum, dumstr, MAXLEN);
         strcpy(text, strdum);
      }
      idum = iinstr((unsigned char*) text, (unsigned char*) "$POINT$(", 0);
   }

   idum = iinstr((unsigned char*) text, (unsigned char*) "$HEX$(", 0);
   while (idum != -1) {
      idum2 = iinstr((unsigned char*) text, (unsigned char*) ")", idum);
      if (idum2 != -1) {
         mid(text, idum + 6, idum2 - 6 - idum, dumstr);
         ltrim(dumstr);
         rtrim(dumstr);
         sscanf(dumstr, "%ld", &ldum);
         sprintf(dumstr, "%04lX", ldum);
         mid(text, 0, idum, strdum);
         strncat(strdum, dumstr, MAXLEN);
         mid(text, idum2 + 1, strlen(text) - idum2 - 1, dumstr);
         strncat(strdum, dumstr, MAXLEN);
         strcpy(text, strdum);
      }
      idum = iinstr((unsigned char*) text, (unsigned char*) "$HEX$(", 0);
   }

   idum = iinstr((unsigned char*) text, (unsigned char*) "$CRC$(", 0);
   while (idum != -1) {
      idum2 = iinstr((unsigned char*) text, (unsigned char*) ")", idum);
      if (idum2 != -1) {
         mid(text, idum + 6, idum2 - 6 - idum, dumstr);
         ltrim(dumstr);
         rtrim(dumstr);
         ldum = strlen(dumstr);
         crc16_init(&crc);
         crc16_process(&crc, (unsigned char*) dumstr, ldum);
         crc16_finalize(&crc);
         while (crc >= 32768) crc -= 32768;
         sprintf(dumstr, "%hu", crc);
         mid(text, 0, idum, strdum);
         strncat(strdum, dumstr, MAXLEN);
         mid(text, idum2 + 1, strlen(text) - idum2 - 1, dumstr);
         strncat(strdum, dumstr, MAXLEN);
         strcpy(text, strdum);
      }
      idum = iinstr((unsigned char*) text, (unsigned char*) "$CRC$(", 0);
   }
}

int execute(char* programm) {
   char** command;

   char   logstr[MAXLEN] = "\0";

   int    anz = 1;
   int    pos = 0;
   int    altpos = -1;
   int    returncode = 0;

   int    idum = 0;

#ifdef __deutsch
   strcpy(logstr, "  Starte ");
#else
   strcpy(logstr, "  Starting ");
#endif
   if (iinstr((unsigned char*) programm, (unsigned char*) ".EXE", 0) == -1 && iinstr((unsigned char*) programm, (unsigned char*) ".COM", 0) == -1) {
      strncat(logstr, programm, MAXLEN - 1);
      writelog("+", logstr);
      returncode = system(programm);
   }
#ifndef __linux__
   else {
      while (idum != -1) {
         idum = iinstr((unsigned char*) programm, (unsigned char*) " ", idum + 1);
         anz++;
      }
      command = new char*[anz];
      for (idum = 0; idum < anz; idum++) {
         command[idum] = new char [50];
#ifdef __deutsch
         if (command[idum] == NULL) writelog ("!", "  Nicht genug Speicher!");
#else
         if (command[idum] == NULL) writelog ("!", "  Not enough memory!");
#endif
      }
      for (idum = 0; idum < anz - 1; idum++) {
         pos = iinstr((unsigned char*) programm, (unsigned char*) " ", altpos + 1);
         if (pos != -1) mid(programm, altpos + 1, pos - altpos - 1, command[idum]);
         else mid(programm, altpos + 1, strlen(programm) - altpos - 1, command[idum]);
         strncat(logstr, command[idum], MAXLEN - 2);
         strncat(logstr, " ", MAXLEN - 1);
         altpos = pos;
      }
      writelog("+", logstr);
      command[anz - 1] = (char*) NULL;
      returncode = spawnvp(P_WAIT | P_DEFAULT, command[0], command);
      for (idum = 0; idum < anz; idum++) delete [] command[idum];
   }
#endif
   if (returncode != -1) {
      sprintf(logstr, "  Returncode: %i", returncode);
      writelog("-", logstr);
   }
#ifdef __deutsch
   else writelog("!", "  Programm nicht gefunden.");
#else
   else writelog("!", "  program not found.");
#endif
   return(returncode);
}

#ifdef __linux__

int strnicmp(char *s1, char *s2, int n) {
   strncasecmp(s1,s2, n);
}

int stricmp(char *s1, char *s2) {
   strcasecmp(s1,s2);
}

#endif

void readmacros(char** argtext, char** pwd, unsigned int* nr) {
   FILE* MACRO = NULL;
   FILE* MACRO_FILE = NULL;

   char dumstr[MAXLEN] = "\0";
   char key_delim[10] = "=\0";
   char keyword[MAXLEN] = "\0";
   char macro_file[255] = "\0";
   char old_keyword[MAXLEN] = "\0";
   char zeile[MAXLEN] = "\0";

   int aktrecord = 0;
   int gefunden = 0;
   int idum = 0;
   int idum2 = 0;
   int keylen = 0;

#ifdef __linux__
   glob_t globdata;
#else
   _find filepointer;
#endif

   if ((MACRO = fopen(macroFile, "rt")) == NULL) {
#ifdef __deutsch
      sprintf(dumstr, "- Kann Macro-Datei '%s' nicht ffnen.", macroFile);
#else
      sprintf(dumstr, "- Can't open macro file '%s'.", macroFile);
#endif
      writelog("!", dumstr);
      printf("\n%s.", dumstr);
   }
   else {
      while (feof(MACRO) == 0) {
         fgets(zeile, MAXLEN, MACRO);
         ltrim(zeile);
         // CRLF abschneiden
         idum = strlen(zeile) - 1;
         if (zeile[idum] == 10) zeile[idum] = 0;
         idum = strlen(zeile) - 1;
         if (zeile[idum] == 13) zeile[idum] = 0;
         if (zeile[0] != ';' && zeile[0] != 0) {
#ifdef __deutsch
            if (strnicmp(zeile, "Datei ", 6) == 0) {
#else
            if (strnicmp(zeile, "File ", 5) == 0) {
#endif
               if (macro_file[0] != 0) {
                  fclose(MACRO_FILE);
                  macro_file[0] = 0;
               }
#ifdef __deutsch
               mid(zeile, 6, strlen(zeile) - 6, macro_file);
#else
               mid(zeile, 5, strlen(zeile) - 5, macro_file);
#endif
               ltrim(macro_file);
               // Makros im Dateinamen bearbeiten
               domacros(macro_file, pwd, *nr, argtext);
               // Evtl. Wildcards im Filenamen abarbeiten
               dumstr[0] = 0;
               idum = -1;
               idum2 = 0;
               idum = iinstr((unsigned char*) macro_file, (unsigned char*) PATH_DELIM, idum + 1);
               while (idum != -1) {
                  idum2 = idum;
                  idum = iinstr((unsigned char*) macro_file, (unsigned char*) PATH_DELIM, idum + 1);
               }
               if (idum2 != 0) mid(macro_file, 0, idum2 + 1, dumstr);
#ifdef __linux__
               idum = glob(macro_file, GLOB_NOSORT, NULL, &globdata);
               if ((idum != 0) || (globdata.gl_pathc == 0)) {
#else
               idum = __findfirst(macro_file, 0, &filepointer);
               if (idum != 0) {
#endif
#ifdef __deutsch
                  sprintf(dumstr, "  Keine zu '%s' passende Datei gefunden.", macro_file);
#else
                  sprintf(dumstr, "  found no files matching '%s'.", macro_file);
#endif
                  writelog("?", dumstr);
                  writelog("!", iGNORE);
                  macro_file[0] = 0;
               }
               else {
                  strcpy(macro_file, dumstr);
#ifdef __linux__
                  strcat(macro_file, globdata.gl_pathv[0]);
#else
                  strcat(macro_file, filepointer.name);
#endif
               }
#ifdef __linux__
               globfree(&globdata);
#endif
#ifdef __deutsch
               sprintf(dumstr, "- Durchsuche nach Macros: '%s'", macro_file);
#else
               sprintf(dumstr, "- Looking for macros in '%s'", macro_file);
#endif
               writelog("+", dumstr);
               if ((MACRO_FILE = fopen(macro_file, "rt")) == NULL) {
                  writelog("?", cANTOPEN);
                  writelog("!", iGNORE);
                  macro_file[0] = 0;
               }
            }
#ifdef __deutsch
            else if (strnicmp(zeile, "Trennung ", tRENNUNGLEN) == 0) {
#else
            else if (strnicmp(zeile, "Separator ", tRENNUNGLEN) == 0) {
#endif
               mid(zeile, tRENNUNGLEN, strlen(zeile) - tRENNUNGLEN, key_delim);
               ltrim(key_delim);
               rtrim(key_delim);
               if (key_delim[0] = '"') mid(key_delim, 1, strlen(key_delim) - 1, key_delim);
               idum = strlen(key_delim) - 1;
               if (key_delim[idum] = '"') key_delim[idum] = 0;
#ifdef __deutsch
               sprintf(dumstr, "Neues Trennzeichen fr Schlsselwrter: '%s'", key_delim);
#else
               sprintf(dumstr, "New seperator for keywords: '%s'", key_delim);
#endif
               writelog("+", dumstr);
            }
            else if (strnicmp(zeile, "$PWD", 4) == 0 && (idum = iinstr((unsigned char*) zeile, (unsigned char*) "$=", 4)) != -1) {
               sprintf(dumstr, bEARBEITEZEILE, zeile);
               writelog("-", dumstr);
               if (macro_file[0] == 0) writelog("!", kEINEAKTMDATEI);
               else {
                  strcpy(old_keyword, keyword);
                  mid(zeile, 7, strlen(zeile) - 7, keyword);
                  ltrim(keyword);
                  rtrim(keyword);
                  strncat(keyword, key_delim, MAXLEN - 1);
                  keylen = strlen(keyword);
                  sscanf(zeile, "$PWD%i$=", &aktrecord);
                  if (aktrecord < 1 || aktrecord > 5) {
#ifdef __deutsch
                     sprintf(dumstr, "  - Zeile enthlt Verweis auf $PWD%i$", aktrecord);
                     writelog("?", dumstr);
                     writelog("!", "    Gltiger Wertebereich $PWD1$ bis $PWD4$, Zeile ignoriert");
#else
                     sprintf(dumstr, "  - line contains $PWD%i$", aktrecord);
                     writelog("?", dumstr);
                     writelog("!", "    valid value range: $PWD1$ to $PWD4$, line ignored");
#endif
                  }
                  else {
#ifdef __deutsch
                     sprintf(dumstr, "  - Suche nach '%s', speichere in $PWD%i$", keyword, aktrecord);
#else
                     sprintf(dumstr, "  - looking for '%s', saving in $PWD%i$", keyword, aktrecord);
#endif
                     writelog("-", dumstr);
                     aktrecord--;
                     gefunden = 0;
                     if (stricmp(keyword, old_keyword) != 0) fseek(MACRO_FILE, 0, SEEK_SET);
                     while (feof(MACRO_FILE) == 0) {
                        fgets(zeile, MAXLEN, MACRO_FILE);
                        ltrim(zeile);
                        // CRLF abschneiden
                        idum = strlen(zeile) - 1;
                        if (zeile[idum] == 10) zeile[idum] = 0;
                        idum = strlen(zeile) - 1;
                        if (zeile[idum] == 13) zeile[idum] = 0;
                        if (strnicmp(zeile, keyword, keylen) == 0) {
                           mid(zeile, keylen, strlen(zeile) - keylen, pwd[aktrecord]);
                           gefunden = 1;
#ifdef __deutsch
                           sprintf(dumstr, "  - Speichere '%s'", pwd[aktrecord]);
#else
                           sprintf(dumstr, "  - Saving '%s'", pwd[aktrecord]);
#endif
                           writelog("+", dumstr);
                           break;
                        }
                     }
                     if (gefunden == 0) writelog("!", sUCHBEGRNOTFOUND);
                  }
               }
            }
            else if (strnicmp(zeile, "$PNT$", 5) == 0) {
               sprintf(dumstr, bEARBEITEZEILE, zeile);
               writelog("-", dumstr);
               if (macro_file[0] == 0) writelog("!", kEINEAKTMDATEI);
               else {
                  strcpy(old_keyword, keyword);
                  mid(zeile, 6, strlen(zeile) - 6, keyword);
                  ltrim(keyword);
                  rtrim(keyword);
                  strncat(keyword, key_delim, MAXLEN - 1);
                  keylen = strlen(keyword);
#ifdef __deutsch
                  sprintf(dumstr, "  - Suche nach '%s', speichere in $PNT$", keyword);
#else
                  sprintf(dumstr, "  - looking for '%s', saving in $PNT$", keyword);
#endif
                  writelog("-", dumstr);
                  gefunden = 0;
                  if (stricmp(keyword, old_keyword) != 0) fseek(MACRO_FILE, 0, SEEK_SET);
                  while (feof(MACRO_FILE) == 0) {
                     fgets(zeile, MAXLEN, MACRO_FILE);
                     ltrim(zeile);
                     // CRLF abschneiden
                     idum = strlen(zeile) - 1;
                     if (zeile[idum] == 10) zeile[idum] = 0;
                     idum = strlen(zeile) - 1;
                     if (zeile[idum] == 13) zeile[idum] = 0;
                     if (strnicmp(zeile, keyword, keylen) == 0) {
                        mid(zeile, keylen, strlen(zeile) - keylen, dumstr);
                        sscanf(dumstr, "%u", nr);
                        gefunden = 1;
#ifdef __deutsch
                        sprintf(dumstr, "  - Speichere %u", *nr);
#else
                        sprintf(dumstr, "  - saving %u", *nr);
#endif
                        writelog("+", dumstr);
                        break;
                     }
                  }
                  if (gefunden == 0) writelog("!", sUCHBEGRNOTFOUND);
               }
            }
            else if (strnicmp(zeile, "$N", 2) == 0 && (idum = iinstr((unsigned char*) zeile, (unsigned char*) "$=", 3)) != -1) {
               sprintf(dumstr, bEARBEITEZEILE, zeile);
               writelog("-", dumstr);
               if (macro_file[0] == 0) writelog("!", kEINEAKTMDATEI);
               else {
                  strcpy(old_keyword, keyword);
                  mid(zeile, idum + 2, strlen(zeile) - idum - 2, keyword);
                  ltrim(keyword);
                  rtrim(keyword);
                  strncat(keyword, key_delim, MAXLEN);
                  keylen = strlen(keyword);
                  sscanf(zeile, "$N%i$=", &aktrecord);
                  if (aktrecord < 1 || aktrecord > MAX) {
#ifdef __deutsch
                     sprintf(dumstr, "  - Zeile enthlt Verweis auf $N%i$", aktrecord);
                     writelog("?", dumstr);
                     sprintf(dumstr, "    Gltiger Wertebereich $N1$ bis $N%i$, Zeile ignoriert", MAX);
#else
                     sprintf(dumstr, "  - line contains $N%i$", aktrecord);
                     writelog("?", dumstr);
                     sprintf(dumstr, "    valid value range: $N1$ to $N%i$, line ignored", MAX);
#endif
                     writelog("!", dumstr);
                  }
                  else {
#ifdef __deutsch
                     sprintf(dumstr, "  - Suche nach '%s', speichere in $N%i$", keyword, aktrecord);
#else
                     sprintf(dumstr, "  - looking for '%s', saving in $N%i$", keyword, aktrecord);
#endif
                     writelog("-", dumstr);
                     gefunden = 0;
                     if (stricmp(keyword, old_keyword) != 0) fseek(MACRO_FILE, 0, SEEK_SET);
                     while (feof(MACRO_FILE) == 0) {
                        fgets(zeile, MAXLEN, MACRO_FILE);
                        ltrim(zeile);
                        // CRLF abschneiden
                        idum = strlen(zeile) - 1;
                        if (zeile[idum] == 10) zeile[idum] = 0;
                        idum = strlen(zeile) - 1;
                        if (zeile[idum] == 13) zeile[idum] = 0;
                        if (strnicmp(zeile, keyword, keylen) == 0) {
                           idum = strlen(zeile);
                           mid(zeile, keylen, idum - keylen, argtext[aktrecord]);
                           gefunden = 1;
#ifdef __deutsch
                           sprintf(dumstr, "  - Speichere '%s'", argtext[aktrecord]);
#else
                           sprintf(dumstr, "  - Saving '%s'", argtext[aktrecord]);
#endif
                           writelog("+", dumstr);
                           break;
                        }
                     }
                     if (gefunden == 0) {
                        argtext[aktrecord][0] = 0;
                        writelog("!", sUCHBEGRNOTFOUND);
                     }
                  }
               }
            }
            else {
#ifdef __deutsch
               sprintf(dumstr, "  - Unbekannte Anweisung: '%s'", zeile);
#else
               sprintf(dumstr, "  - unknown command: '%s'", zeile);
#endif
               writelog("!", dumstr);
               writelog("!", iGNORE);
            }
         }
      }
      if (macro_file[0] != 0) fclose(MACRO_FILE);
   }
}

int move(char* source, char* dest) {
   FILE* src;
   FILE* dst;

   char* buffer;
   int   charloaded = 0;

   buffer = (char*) calloc(4096, 1);
   if (buffer == NULL) return(-1);
   if ((src = fopen(source, "rb")) != NULL) {
      if ((dst = fopen(dest, "wb")) != NULL) {
         while(feof(src) == 0) {
            charloaded = fread(buffer, 1, 4096, src);
            charloaded = fwrite(buffer, 1, charloaded, dst);
         }
         fclose(src);
         fclose(dst);
         remove(source);
         free(buffer);
         return(0);
      }
      fclose(src);
   }
   return(-1);
}

int main (int argc, char*argv[]) {
   char** argstuff;
   char** pwd;

   char dumstr[MAXLEN] = "\0";
   char ext_datei[255] = "\0";
   char newline[MAXLEN] = "\0";
   char part1[255] = "\0";
   char resultpnt[255] = "\0";
   char resultpwd[255] = "\0";
   char strdum[MAXLEN] = "\0";
#ifdef __deutsch
   char textmarke[255] = "; +++ FREEPNT-MARKE +++";
#else
   char textmarke[255] = "; +++ FREEPNT-LABEL +++";
#endif
   char tmp_datei[255] = "\0";
   char zeile[MAXLEN] = "\0";

   FILE* AUTO = NULL;
   FILE* EXT = NULL;
   FILE* CONFIG = NULL;
   FILE* TMP = NULL;

   unsigned int* nrn;
   unsigned int  nr = 0;

   time_t systemtime = 0;

   int* block_status;

   int akt_arg = 0;
   int akt_block = 0;
   int akt_pwd = 0;
   int anzahl = 0;
   int autoconfig = 0;
   int fail = 0;
   int file_mode = 0;
   int idum = 0;
   int idum2 = 0;
   int idum3 = 0;
   int newline_written = 0;
   int oper = 0;
   int read_macros = 0;
   int screen = 0;
   int textmarkelen = 23;
   int writemode = 0;

   printf("\n%s, (c)1999-2001 Christian von Busse, 2:240/2188, keks@kruemel.org", VERSION);
   printf("\n                             Released under the GNU public license, see COPYING\n");
#ifdef __deutsch
   printf("\n              nderungen fr Linux von Matthias Tichy, 2:2432/645, mtt@tichy.de\n");
#else
   printf("\n                  Changes for Linux by Matthias Tichy, 2:2432/645, mtt@tichy.de\n");
#endif

   configFile = (char *) malloc(12 + 1);
   strcpy(configFile, "freepnt.cfg");

   macroFile = (char *) malloc(12 + 1);
   strcpy(macroFile, "freepnt.mac");

   // LOGFile-Namen als erstes ermitteln
   // (Kommandzeilenparameter "-l")
   for (idum = 1; idum < argc; idum++) if (strnicmp(argv[idum], "-l:", 3) == 0) {
      mid(argv[idum], 3, strlen(argv[idum]) - 3, logfile);
      break;
   }

   // Hallo! im Log
#ifdef __deutsch
   sprintf(dumstr, "%s, Programmstart", VERSION);
#else
   sprintf(dumstr, "%s, program launched", VERSION);
#endif
   writelog("+", dumstr);

   argstuff =  (char**) malloc(sizeof(char *) * MAX);
   for (idum = 1; idum < MAX; idum++) argstuff[idum] = (char*) malloc(255);

   pwd =  (char**) malloc(sizeof(char *) * 4);
   for (idum = 0; idum < 4; idum++) {
      pwd[idum] = (char*) malloc(9);
      for (idum2 = 0; idum2 < 9; idum2++) pwd[idum][idum2] = 0;
   }

   for (akt_arg = 1; akt_arg < argc; akt_arg++) {
      if (strnicmp(argv[akt_arg], "-cfg:", 5) == 0) {
         mid(argv[akt_arg], 5, strlen(argv[akt_arg]) - 5, dumstr);
         free(configFile);
         configFile = (char *) malloc(strlen(dumstr)+1);
         strcpy(configFile, dumstr);
#ifdef __deutsch
         sprintf(dumstr, "Alternative Config vorgegeben: '%s'", configFile);
#else
         sprintf(dumstr, "new config selected: '%s'", configFile);
#endif
         writelog("+", dumstr);
      }
      else if (strnicmp(argv[akt_arg], "-macro:", 7) == 0) {
         mid(argv[akt_arg], 7, strlen(argv[akt_arg]) - 7, dumstr);
         free(macroFile);
         macroFile = (char *) malloc(strlen(dumstr)+1);
         strcpy(macroFile, dumstr);
#ifdef __deutsch
         sprintf(dumstr, "Alternative Macro-Datei vorgegeben: '%s'", macroFile);
#else
         sprintf(dumstr, "new macro file selected: '%s'", macroFile);
#endif
         writelog("+", dumstr);
      }
      else if (strnicmp(argv[akt_arg], "-c", 2) == 0) autoconfig = 1;
      else if (strnicmp(argv[akt_arg], "-m", 2) == 0) read_macros = 1;
      else if (strnicmp(argv[akt_arg], "-pwd:", 5) == 0) {
         mid(argv[akt_arg], 5, strlen(argv[akt_arg]) - 5, pwd[0]);
         for (idum = 1; idum < 4; idum++) strcpy(pwd[idum], pwd[0]);
#ifdef __deutsch
         sprintf(dumstr, "Passwrter vorgegeben: '%s'", pwd[0]);
#else
         sprintf(dumstr, "passwords selected: '%s'", pwd[0]);
#endif

         writelog("+", dumstr);
      }
      else if (strnicmp(argv[akt_arg], "-pnt:", 5) == 0) {
         mid(argv[akt_arg], 5, strlen(argv[akt_arg]) - 5, dumstr);
         idum = iinstr((unsigned char*) dumstr, (unsigned char*) ".", 0);
         if (idum != -1) mid(dumstr, idum + 1, strlen(dumstr) - idum - 1, dumstr);
         sscanf(dumstr, "%u", &nr);
#ifdef __deutsch
         sprintf(dumstr, "Point-Nr. vorgegeben: '%u'", nr);
#else
         sprintf(dumstr, "point number selected: '%u'", nr);
#endif
         writelog("+", dumstr);
      }
      else if (strnicmp(argv[akt_arg], "-scrn", 5) == 0) {
#ifdef __deutsch
         writelog("+", "Ausgabe erfolgt am Bildschirm.");
         printf("\n%s.\nAusgabe erfolgt am Bildschirm!",dumstr);
#else
         writelog("+", "output to screen");
         printf("\n%s.\nOutput to screen",dumstr);
#endif
         screen = 1;
      }
      else if (strnicmp(argv[akt_arg], "-n", 2) == 0 && ((idum3 = iinstr((unsigned char*) argv[akt_arg], (unsigned char*) ":", 0)) != -1)) {
         sscanf(argv[akt_arg], "-n%i:", &idum2);
         if (idum2 > 0 && idum2 < 26) {
            mid(argv[akt_arg], idum3 + 1, strlen(argv[akt_arg]) - idum3 - 1, argstuff[idum2]);
#ifdef __deutsch
            sprintf(dumstr, "Speichere '%s' als $N%i$", argstuff[idum2], idum2);
#else
            sprintf(dumstr, "saving '%s' as $N%i$", argstuff[idum2], idum2);
#endif
            writelog("+", dumstr);
         }
         else {
#ifdef __deutsch
            sprintf(dumstr, "Ungltiger '-n'-Parameter: '%s'", argv[akt_arg]);
            writelog("!", dumstr);
            sprintf(dumstr, "Die Werte fr -n reichen von -n1 bis -n%i", MAX);
            writelog("!", dumstr);
#else
            sprintf(dumstr, "invalid -n parameter: '%s'", argv[akt_arg]);
            writelog("!", dumstr);
            sprintf(dumstr, "valid values range from -n1 to -n%i", MAX);
            writelog("!", dumstr);
#endif
         }
      }
      else {
         // Parameter -l: ignorieren
         if (strnicmp(argv[akt_arg], "-l:", 3) != 0) {
#ifdef __deutsch
            sprintf(dumstr, "Unbekannter Parameter '%s' ignoriert.", argv[akt_arg]);
#else
            sprintf(dumstr, "unknown parameter '%s' ignored.", argv[akt_arg]);
#endif

            writelog("?", dumstr);
         }
      }
   }

   if (screen == 0 && autoconfig == 0 && read_macros == 0) {
#ifdef __deutsch
      printf("\nKommandozeilenparameter:\n");
      printf("\n-c      Automatische Konfiguration anhand von FREEPNT.CFG");
      printf("\n-cfg:   ndern des Namens von FREEPNT.CFG");
      printf("\n-m      Makros einlesen aus FREEPNT.MAC");
      printf("\n-macro: ndern des Namens von FREEPNT.MAC");
      printf("\n-nx:    Makro $Nx$ aus Kommandozeile auslesen.");
      printf("\n-pnt:   Pointnummer aus Kommandozeile auslesen.");
      printf("\n-pwd:   Passwort aus der Kommandozeile auslesen.");
      printf("\n-l:     Namen des LOGFiles (FREEPNT.LOG) ndern.");
      printf("\n-scrn   Debug-Modus, Ergebnisse werden auf dem Bildschirm ausgegeben.\n");
      strcpy(dumstr, "Keine Aktionen auszufhren, Programm beendet.");
#else
      printf("\nCommand line parameters:\n");
      printf("\n-c      automatic configuration as specified in FREEPNT.CFG");
      printf("\n-cfg:   change the name of FREEPNT.CFG");
      printf("\n-m      read macros from FREEPNT.MAC");
      printf("\n-macro: change the name of FREEPNT.MAC");
      printf("\n-nx:    read macro $Nx$ from the command line.");
      printf("\n-pnt:   read pointnumber from the command line.");
      printf("\n-pwd:   read passwords from the command line.");
      printf("\n-l:     change the name of the logfile (FREEPNT.LOG).");
      printf("\n-scrn   debug mode, all outputs to the screen.\n");
      strcpy(dumstr, "nothing to do, program end.");
#endif
      writelog("-", dumstr);
      printf("\n%s\n", dumstr);
      exit(0);
   }

   if (read_macros == 1) {
#ifdef __deutsch
      writelog("+", "Lade Macros.");
#else
      writelog("+", "Loading macros.");
#endif
      readmacros(argstuff, pwd, &nr);
   }

   if (autoconfig == 1) {
      if ((AUTO = fopen(configFile, "rt")) == NULL) {
#ifdef __deutsch
         printf("\n%s kann nicht im aktuellen Verzeichnis geffnert werden.\nKommandozeilenparameter -c wird ignoriert.\n", configFile);
         writelog("!", "Die Configdatei kann nicht geffnet werden, -c ignoriert.");
#else
         printf("\n%s cannot be opened in the current directory.\nCommand line parameter -c ignored.\n", configFile);
         writelog("!", "can't open config file, -c ignored.");
#endif
         autoconfig = 0;
      }
#ifdef __deutsch
      else writelog("+", "Autokonfiguration aktiviert.");
#else
      else writelog("+", "automatic configuration activated.");
#endif
   }

   // Nchste freie Pointnummer nur ermitteln, wenn keine vorgegeben wurde
   if (nr == 0) {
      if ((CONFIG = fopen("FREEPNT.TXT", "rt")) == NULL) {
#ifdef __deutsch
         printf("\n\nFREEPNT.TXT kann nicht im aktuellen Verzeichnis geffnet werden.\nProgrammende (255).");
         writelog("!", "FREEPNT.TXT nicht gefunden, Programmende (255)");
#else
         printf("\n\nFREEPNT.TXT cannot be opened in the current directory.\nEnd (255).");
         writelog("!", "FREEPNT.TXT not found, end (255)");
#endif
         exit(255);
      }
      while (feof(CONFIG) == 0) {
         fgets(zeile, 80, CONFIG);
         if (zeile[0] != 0) anzahl++;
      }
#ifdef __deutsch
      sprintf(dumstr, "FREEPNT.TXT enthlt %i vergebene Pointnummern.", anzahl);
#else
      sprintf(dumstr, "FREEPNT.TXT contains %i pointnumbers already in use.", anzahl);
#endif
      writelog("-", dumstr);
      fseek(CONFIG, 0, SEEK_SET);
      nrn = (unsigned int*) malloc (sizeof (unsigned int) * anzahl);
      for (idum = 0; idum < anzahl; idum++) {
         fgets(zeile, 80, CONFIG);
         if (zeile[0] != 0) sscanf(zeile, "%u", &nrn[idum]);
      }
      fclose(CONFIG);
      for (nr = 1; nr <= 65535; nr++) {
         fail = 0;
         for (idum = 0; idum < anzahl; idum++) {
            if (nrn[idum] == nr) {
               fail = 1;
               break;
            }
         }
         if (fail == 0) break;
      }
      if (fail == 1) {
#ifdef __deutsch
         printf("\n\nEs konnte keine neue Pointnummer vergeben werden.\nAlle Pointnummern sind belegt!");
         writelog("!", "Keine neue Pointnummer vergeben, alle belegt!");
         writelog("!", "Programmende (253)");
#else
         printf("\n\nNo new pointnumber could be issued, all are in use!");
         writelog("!", "no new pointnumber issued, all are in use!");
         writelog("!", "end (253)");
#endif
         exit(253);
      }
      else if (screen == 0) {
         if ((CONFIG = fopen("FREEPNT.TXT", "at")) != NULL) {
            sprintf(dumstr,"%u\0", nr);
            fputs(dumstr, CONFIG);
            fputc(10, CONFIG);
            fclose(CONFIG);
         }
         else {
#ifdef __deutsch
            printf("\nIn FREEPNT.TXT konnte nicht geschrieben werden!");
            writelog("!", "In FREEPNT.TXT konnte nicht geschrieben werden!");
            writelog("!", "FREEPNT.TXT konnte nicht aktualisiert werden!");
            writelog("!", "Programmende (254)");
#else
            printf("\nCannot save FREEPNT.TXT!");
            writelog("!", "cannot save FREEPNT.TXT!");
            writelog("!", "FREEPNT.TXT was not updated!");
            writelog("!", "end (254)");
#endif
            exit(254);
         }
      }
   }

   systemtime = time(NULL);
   while (systemtime > 65535) systemtime -= (long unsigned int) rint(systemtime / 65535) * 65535;
   srand(systemtime);

   // Passwrter nur heraussuchen, wenn noch keines angegeben ist
   for (akt_pwd = 0; akt_pwd < 4; akt_pwd++) {
      if (pwd[akt_pwd][0] == 0) {
         for (idum = 0; idum < 8; idum++) {
            idum2 = (rand() % 36) - 10; // 36 Zeichen
            if (idum2 < 0) pwd[akt_pwd][idum] = idum2 + 58;
            else pwd[akt_pwd][idum] = idum2 + 65;
         }
      }
      // Evtl. Standard-Passwort verlangt?
      else if (strnicmp(pwd[akt_pwd], "std", 5) == 0) {
         sprintf(pwd[akt_pwd], "PNT%u", nr);
#ifdef __deutsch
         sprintf(dumstr, "Vergebe Standardpasswort Nr. %i fr neuen Point", akt_pwd);
#else
         sprintf(dumstr, "issuing default password Nr. %i for new point", akt_pwd);
#endif

         writelog("+", dumstr);
      }
   }

   // Ergebnisse ausgeben:
#ifdef __deutsch
   sprintf(dumstr, "Ergebnis: Point-Nr.: %u, PWD: %s, %s, %s, %s", nr, pwd[0], pwd[1], pwd[2], pwd[3]);
#else
   sprintf(dumstr, "Result: Pointnumber: %u, PWD: %s, %s, %s, %s", nr, pwd[0], pwd[1], pwd[2], pwd[3]);
#endif
   writelog("+", dumstr);
   if (screen == 1) printf("\n\n%s\n", dumstr);

   // Autokonfiguration?
   if (autoconfig == 1) {
#ifdef __deutsch
      writelog("+", "Fhre Autokonfiguration durch");
#else
      writelog("+", "Executing automatic configuration");
#endif

      // IF-Blcke vorbereiten
      block_status =  (int*) malloc(sizeof(int) * 10);
      for (idum = 0; idum < 10; idum++) block_status[idum] = 0;
      akt_block = 0;

      while (feof(AUTO) == 0) {
         fgets(zeile, MAXLEN, AUTO);
         if (feof(AUTO) != 0) break;
         ltrim(zeile);
         // CRLF abschneiden
         idum = strlen(zeile) - 1;
         if (zeile[idum] == 10) zeile[idum] = 0;
         idum = strlen(zeile) - 1;
         if (zeile[idum] == 13) zeile[idum] = 0;
         if (zeile[0] != ';' && zeile[0] != '#') {
#ifdef __deutsch
            if (strnicmp(zeile, "Ende Wenn", 9) == 0) {
#else
            if (strnicmp(zeile, "End If", 6) == 0) {
#endif
               if (akt_block > 0) {
                  block_status[akt_block] = 0;
                  akt_block--;
               }
#ifdef __deutsch
               else writelog("!", "'Ende Wenn' ohne zuzuordnendes Wenn!");
#else
               else writelog("!", "'End If' without corresponding If!");
#endif
            }
            if (block_status[akt_block] == 0) {
               // Kommandos im aktuellen Block bearbeiten
#ifdef __deutsch
               if (strnicmp(zeile, "Datei ", 6) == 0) {
#else
               if (strnicmp(zeile, "File ", 5) == 0) {
#endif
                  if (ext_datei[0] != 0) fclose(EXT);
                  if (tmp_datei[0] != 0) {
                     fclose(TMP);
                     if (screen == 1) remove(tmp_datei);
                     else {
                        remove(ext_datei);
                        idum = rename(tmp_datei, ext_datei);
                        if (idum == -1) {
                           sprintf(dumstr, uMBENENNENERR, tmp_datei, ext_datei);
                           writelog("!", dumstr);
                        }
                     }
                     tmp_datei[0] = 0;
                  }
                  file_mode = 0;
                  sprintf(dumstr, bEARBEITEZEILE, zeile);
                  writelog("+", dumstr);
#ifdef __deutsch
                  mid(zeile, 6, strlen(zeile) - 6, ext_datei);
#else
                  mid(zeile, 5, strlen(zeile) - 5, ext_datei);
#endif
                  domacros(ext_datei, pwd, nr, argstuff);
                  if ((EXT = fopen(ext_datei, "rt")) == NULL) {
                     if ((EXT = fopen(ext_datei, "wt")) == NULL) {
#ifdef __deutsch
                        sprintf(dumstr, "  '%s' kann nicht geffnet werden. Datei-Angabe ignoriert.", ext_datei);
#else
                        sprintf(dumstr, "  can't open '%s'. File specification ignored.", ext_datei);
#endif
                        writelog("!", dumstr);
                        printf("\n%s", dumstr);
                        ext_datei[0] = 0;
                     }
                     else {
                        file_mode = 1;
#ifdef __deutsch
                        sprintf(dumstr, "Erstelle neue Datei '%s'", ext_datei);
#else
                        sprintf(dumstr, "creating new file '%s'", ext_datei);
#endif
                        writelog("-", dumstr);
                        printf("\n%s...", dumstr);
                     }
                  }
                  else {
#ifdef __deutsch
                     sprintf(dumstr, "Bearbeite Datei '%s'", ext_datei);
#else
                     sprintf(dumstr, "Processing file '%s'", ext_datei);
#endif
                     printf("\n%s...", dumstr);
                     writelog("-", dumstr);
                  }
               }
#ifdef __deutsch
               else if (strnicmp(zeile, "Schreibe", 8) == 0 || strnicmp(zeile, "Naechste Zeile", 14) == 0 || strnicmp(zeile, "Ende Zeile", 10) == 0 || strnicmp(zeile, "Anfuegen", 8) == 0) {
                  if (strnicmp(zeile, "Schreibe", 8) == 0) {
                     writemode = 0;
                     idum = 8;
                  }
                  else if (strnicmp(zeile, "Naechste Zeile", 14) == 0){
                      writemode = 1;
                      idum = 14;
                  }
                  else if (strnicmp(zeile, "Ende Zeile", 10) == 0){
                      writemode = 2;
                      idum = 10;
                  }
                  else {
                      writemode = 3;
                      idum = 8;
                  }
#else
               else if (strnicmp(zeile, "Write", 5) == 0 || strnicmp(zeile, "Next Line", 9) == 0 || strnicmp(zeile, "line end", 8) == 0 || strnicmp(zeile, "Append", 6) == 0) {
                  if (strnicmp(zeile, "Write", 5) == 0){
                      writemode = 0;
                      idum = 5;
                  }
                  else if (strnicmp(zeile, "Next Line", 9) == 0) {
                      writemode = 1;
                      idum = 9;
                  }
                  else if (strnicmp(zeile, "Line End", 8) == 0) {
                      writemode = 2;
                      idum = 8;
                  }
                  else {
                      writemode = 3;
                      idum = 6;
                  }
#endif
                  sprintf(dumstr, bEARBEITEZEILE, zeile);
                  writelog("+", dumstr);
                  newline_written = 0;
                  // ggf. 1. Leerzeichen direkt nach 'Schreibe' ignorieren:
                  if (zeile[idum] == 32) mid(zeile, idum + 1, strlen(zeile) - idum - 1, newline);
                  else mid(zeile, idum, strlen(zeile) - idum, newline);
                  domacros(newline, pwd, nr, argstuff);
                  if (ext_datei[0] == 0 || EXT == NULL) {
#ifdef __deutsch
                     writelog("!", "  Keine aktuelle Datei gewhlt.");
                     sprintf(dumstr, "  Zeile nicht ausgefhrt.");
#else
                     writelog("!", "  No current file selected.");
                     sprintf(dumstr, "  line not processed.");
#endif
                     writelog("!", dumstr);
                  }
                  else {
                     if (tmp_datei[0] == 0) {
                        idum = -1;
                        do {
                           idum2 = idum;
                           idum = iinstr((unsigned char*) ext_datei, (unsigned char*) PATH_DELIM, idum2 + 1);
                        }
                        while (idum != -1);
                        mid(ext_datei, 0, idum2 + 1, tmp_datei);
                        strncat(tmp_datei, "freepnt.$$$", 255);
                        if ((TMP = fopen(tmp_datei, "wt")) == NULL) {
#ifdef __deutsch
                           sprintf(dumstr, "  Kann Temporr-Datei '%s' nicht erstellen. Zeile nicht geschrieben.", tmp_datei);
#else
                           sprintf(dumstr, "  can't create temp file '%s'. Line not written.", tmp_datei);
#endif
                           writelog("!", dumstr);
                        }
                     }
#ifdef __deutsch
                     sprintf(dumstr, "  Neue Zeile: '%s'", newline);
#else
                     sprintf(dumstr, "  new line: '%s'", newline);
#endif
                     writelog("+", dumstr);
#ifdef __deutsch
                     if (screen == 1) printf("\nNeue Zeile: '%s'...", newline);
#else
                     if (screen == 1) printf("\nNew line: '%s'...", newline);
#endif
                     if (file_mode == 0) {
                        fseek(EXT, 0, SEEK_SET);
                        while (feof(EXT) == 0) {
                           fgets(dumstr, MAXLEN, EXT);
                           // Text-Marke fr's Einfgen suchen, falls vorhanden
                           if (strnicmp(dumstr, textmarke, textmarkelen) == 0) {
                              switch (writemode) {
                                 case 0:
                                    fputs(newline, TMP);
                                    fputc(10, TMP);
                                    newline_written = 1;
#ifdef __deutsch
                                    writelog("-", "  ...vor Textmarke eingefgt.");
                                    if (screen == 1) printf("\nNeue Zeile vor Textmarke eingefgt.");
#else
                                    writelog("-", "  ...inserted in front of label.");
                                    if (screen == 1) printf("\nNew line inserted in front of label.");
#endif
                                    fputs(dumstr, TMP);
                                    break;
                                 case 1:
                                    fputs(dumstr, TMP);
                                    fputs(newline, TMP);
                                    fputc(10, TMP);
                                    newline_written = 1;
#ifdef __deutsch
                                    writelog("-", "  ...in neuer Zeile nach der Textmarke eingefgt.");
                                    if (screen == 1) printf("\nNeue Zeile in der Zeile nach der Textmarke eingefgt.");
#else
                                    writelog("-", "  ...inserted into a new line after label.");
                                    if (screen == 1) printf("\nNew line inserted as new line after the label.");
#endif
                                    break;
                                 case 2:
                                    dumstr[strlen(dumstr) - 1] = 0;
                                    strncat(dumstr, newline, MAXLEN - 1);
                                    fputs(dumstr, TMP);
                                    fputc(10, TMP);
                                    newline_written = 1;
#ifdef __deutsch
                                    writelog("-", "  ...am Ende der Zeile mit der Textmarke eingefgt.");
                                    if (screen == 1) printf("\nNeue Zeile am Ende der Zeile mit der Textmarke eingefgt.");
#else
                                    writelog("-", "  ...appended at the end of the line containing the label.");
                                    if (screen == 1) printf("\nNew line appended at the end of the line containing the label.");
#endif
                                    break;
                                 case 3:
                                    mid(dumstr, 0, textmarkelen, strdum);
                                    strncat(strdum, newline, MAXLEN - 1);
                                    mid(dumstr, textmarkelen, strlen(dumstr) - textmarkelen, dumstr);
                                    strncat(strdum, dumstr, MAXLEN - 1);
                                    fputs(strdum, TMP);
                                    fputc(10, TMP);
                                    newline_written = 1;
#ifdef __deutsch
                                    writelog("-", "  ...direkt hinter der Textmarke eingefgt.");
                                    if (screen == 1) printf("\nNeue Zeile direkt hinter der Textmarke eingefgt.");
#else
                                    writelog("-", "  ...inserted right behind the label.");
                                    if (screen == 1) printf("\nNew line inserted right behind the label.");
#endif
                                    break;
                              }
                           }
                           else fputs(dumstr, TMP);
                        }
                     }
                     if (newline_written == 0) {
                        fputs(newline, TMP);
                        fputc(10, TMP);
#ifdef __deutsch
                        writelog("-", "  Neue Zeile am Ende angefgt.");
                        if (screen == 1) printf("\n...am Ende angefgt.");
#else
                        writelog("-", "  new line append at the end.");
                        if (screen == 1) printf("\n...appended at the end.");
#endif
                        // Ab jetzt neue Zeilen am Ende anfgen
                        file_mode = 1;
                     }
                     if (file_mode == 0) {
                        fclose(TMP);
                        if (screen == 1) remove(tmp_datei);
                        else {
                           fclose(EXT);
                           remove(ext_datei);
                           idum = rename(tmp_datei, ext_datei);
                           if (idum == -1) {
                              sprintf(dumstr, uMBENENNENERR, tmp_datei, ext_datei);
                              writelog("!", dumstr);
                           }
                           EXT = fopen(ext_datei, "rt");
                        }
                        tmp_datei[0] = 0;
                     }
                  }
               }
#ifdef __deutsch
               else if (strnicmp(zeile, "Starte ", sTARTLEN) == 0) {
#else
               else if (strnicmp(zeile, "Start ", sTARTLEN) == 0) {
#endif
                  sprintf(dumstr, bEARBEITEZEILE, zeile);
                  writelog("+", dumstr);
                  if (ext_datei[0] != 0) fclose(EXT);
                  if (tmp_datei[0] != 0) {
                     fclose(TMP);
                     if (screen == 1) remove(tmp_datei);
                     else {
                        remove(ext_datei);
                        idum = rename(tmp_datei, ext_datei);
                        if (idum == -1) {
                           sprintf(dumstr, uMBENENNENERR, tmp_datei, ext_datei);
                           writelog("!", dumstr);
                        }
                     }
                     tmp_datei[0] = 0;
                  }
                  file_mode = 0;
                  ext_datei[0] = 0;
                  mid(zeile, sTARTLEN, strlen(zeile) - sTARTLEN, dumstr);
                  domacros(dumstr, pwd, nr, argstuff);
                  execute(dumstr);
               }
#ifdef __deutsch
               else if (strnicmp(zeile, "Textmarke ", fLAGLEN) == 0) {
#else
               else if (strnicmp(zeile, "Label ", fLAGLEN) == 0) {
#endif
                  sprintf(dumstr, bEARBEITEZEILE, zeile);
                  writelog("+", dumstr);
                  mid(zeile, fLAGLEN, strlen(zeile) - fLAGLEN, textmarke);
                  domacros(textmarke, pwd, nr, argstuff);
                  textmarkelen = strlen(textmarke);
#ifdef __deutsch
                  sprintf(dumstr, "Speichere neue Textmarke '%s' (%i Zeichen)", textmarke, textmarkelen);
#else
                  sprintf(dumstr, "saving new label '%s' (%i chars)", textmarke, textmarkelen);
#endif
                  writelog("+", dumstr);
               }
#ifdef __deutsch
               else if (strnicmp(zeile, "Wenn ", iFLEN) == 0) {
#else
               else if (strnicmp(zeile, "If ", iFLEN) == 0) {
#endif
                  sprintf(dumstr, bEARBEITEZEILE, zeile);
                  writelog("+", dumstr);
                  if (akt_block == 9) {
#ifdef __deutsch
                     writelog("!", "Maximal 9 verschachtelte Blcke sind erlaubt!");
                     writelog("!", "Zeile ignoriert, WENN-Block ausgefhrt.");
#else
                     writelog("!", "No more than 9 interlocked if blocks are allowed!");
                     writelog("!", "Line ignored, if block executed.");
#endif
                  }
                  else {
                     akt_block++;
                     block_status[akt_block] = 1; // Block nicht ausfhren (default)
                     part1[0] = 0;
                     oper = 0;
                     mid(zeile, iFLEN, strlen(zeile) - iFLEN, zeile);
                     ltrim(zeile);
                     domacros(zeile, pwd, nr, argstuff);
                     for (idum = 0; idum < strlen(zeile); idum++) {
                        dumstr[0] = zeile[idum];
                        dumstr[1] = 0;
                        if (iinstr((unsigned char*) "=!><", (unsigned char*) dumstr, 0) != -1) {
                           // ! ist kein Operator, wenn er nicht vor = kommt
                           if ((zeile[idum] == '!' && zeile[idum + 1] == '=') || zeile[idum] != '!') {
                              if (oper == 0) {
                                 mid(zeile, 0, idum, part1);
                                 idum2 = idum;
                              }
                              switch (zeile[idum]) {
                                 case '=':
                                    oper = oper | 1;
                                    break;
                                 case '!':
                                    oper = oper | 2;
                                    break;
                                 case '>':
                                    oper = oper | 4;
                                    break;
                                 case '<':
                                    oper = oper | 8;
                                    break;
                              }
                           }
                        }
                        else if (oper != 0) {
                           mid(zeile, idum, strlen(zeile) - idum, zeile);
                           idum = strlen(zeile);
                        }
                     }
                     if (oper != 0 && zeile[0] != 0) {
                        rtrim(part1);
                        ltrim(zeile);
                        rtrim(zeile);
                        // Jetzt vergleichen
                        if (zeile[0] == '"' && zeile[strlen(zeile) - 1] == '"') {
                           zeile[strlen(zeile) - 1] = 0;
                           mid(zeile, 1, strlen(zeile) - 1, zeile);
                           // Zeichenkettenvergleich
#ifdef __deutsch
                           sprintf(dumstr, "- Vergleiche Zeichenketten '%s' und '%s'", part1, zeile);
#else
                           sprintf(dumstr, "- comparing strings '%s' and '%s'", part1, zeile);
#endif
                           writelog("-", dumstr);
                           // Vergleich immer nur =, Operatoren > und < werden ebenso behandelt
                           if (stricmp(part1, zeile) == 0) block_status[akt_block] = 0;
                        }
                        else {
                           // Wertvergleich
                           sscanf(part1, "%i", &idum);
                           sscanf(zeile, "%i", &idum2);
#ifdef __deutsch
                           sprintf(dumstr, "- Vergleiche Werte %i und %i", idum, idum2);
#else
                           sprintf(dumstr, "- comparing values %i and %i", idum, idum2);
#endif
                           writelog("-", dumstr);
                           if (oper & 1) if (idum == idum2) block_status[akt_block] = 0;
                           if (oper & 4) if (idum > idum2) block_status[akt_block] = 0;
                           if (oper & 8) if (idum < idum2) block_status[akt_block] = 0;
                        }
                        if (oper & 2) {
                           // Mir ist gerade der XOR-Operator entfallen :)
                           if (block_status[akt_block] == 1) block_status[akt_block] = 0;
                           else block_status[akt_block] = 1;
                        }
#ifdef __deutsch
                        if (block_status[akt_block] == 0) writelog("-", "  Bedingung ist erfllt.");
                        else writelog("-", "  Bedingung ist nicht erfllt.");
#else
                        if (block_status[akt_block] == 0) writelog("-", "  condition is matched.");
                        else writelog("-", "  condition is not matching.");
#endif
                     }
#ifdef __deutsch
                     else writelog("!", "  Falsche WENN-Syntax, WENN-Block nicht ausgefhrt.");
#else
                     else writelog("!", "  error in IF syntax, IF block not executed.");
#endif
                  }
               }
#ifdef __deutsch
               else if (strnicmp(zeile, "Ende Wenn", 9) == 0) writelog("-", "- Ende WENN-Block");
#else
               else if (strnicmp(zeile, "End If", 6) == 0) writelog("-", "- end of IF block");
#endif
               else {
#ifdef __deutsch
                  sprintf(dumstr, "Zeile '%s' ignoriert", zeile);
#else
                  sprintf(dumstr, "line '%s' ignored", zeile);
#endif
                  writelog("?", dumstr);
               }
            }
         }
      }
      if (ext_datei[0] != 0) fclose(EXT);
      if (tmp_datei[0] != 0) {
         fclose(TMP);
         if (screen == 1) remove(tmp_datei);
         else {
            remove(ext_datei);
            idum = rename(tmp_datei, ext_datei);
            if (idum == -1) {
               sprintf(dumstr, uMBENENNENERR, tmp_datei, ext_datei);
               writelog("!", dumstr);
            }
         }
      }
   }

   // Programmende
#ifdef __deutsch
   writelog("-", "Programmende (0)");
#else
   writelog("-", "end (0)");
#endif
   writelog("0", "");

}
